@@ -3,6 +3,9 @@
 
 make distclean
 
+# Reconfigure (with force) to get the latest revision from git
+autoreconf -f
+
 if ! ./configure.windows-x86; then
 	exit 1
 fi
@@ -13,7 +16,15 @@ if ! test "x$1" = "x-f"; then
 	fi
 fi
 
-if ! make distwindows distclean; then
+if ! make distwindows-x86 distclean; then
+	exit 1
+fi
+
+if ! ./configure.windows-x64; then
+	exit 1
+fi
+
+if ! make distwindows-x64 distclean; then
 	exit 1
 fi
 
