@@ -456,13 +456,15 @@ string zip_entry::name_get() const
 }
 
 /** Check central directory entry. */
-void zip_entry::check_cent(const unsigned char* buf) const
+void zip_entry::check_cent(const unsigned char* buf, unsigned buf_size) const
 {
+	if (buf_size < ZIP_CO_FIXED) {
+		throw error_invalid() << "Invalid central directory data";
+	}
 	// check signature
 	if (le_uint32_read(buf+ZIP_CO_central_file_header_signature) != ZIP_C_signature) {
 		throw error_invalid() << "Invalid central directory signature";
 	}
-
 	// check filename_length > 0, can't exist a file without a name
 	if (le_uint16_read(buf+ZIP_CO_filename_length) == 0) {
 		throw error_invalid() << "Empty filename in central directory";
@@ -527,13 +529,20 @@ void zip_entry::check_local(const unsigned char* buf) const
 
 void zip_entry::check_descriptor(const unsigned char* buf) const
 {
+	if (0x08074b50 != le_uint32_read(buf+ZIP_DO_header_signature)) {
+		throw error_invalid() << "Invalid header signature on data descriptor " << le_uint32_read(buf+ZIP_DO_crc32);
+	}
 	if (info.crc32 != le_uint32_read(buf+ZIP_DO_crc32)) {
 		throw error_invalid() << "Invalid crc on data descriptor " << info.crc32 << "/" << le_uint32_read(buf+ZIP_DO_crc32);
 	}
-	if (info.compressed_size != le_uint32_read(buf+ZIP_DO_compressed_size)) {
+	if (info.compressed_size != le_uint32_read(buf+ZIP_DO_compressed_size)
+		// allow a 0 size, GNU unzip also allow it
+		&& 0 != le_uint32_read(buf+ZIP_DO_compressed_size)) {
 		throw error_invalid() << "Invalid compressed size in data descriptor " << info.compressed_size << "/" << le_uint32_read(buf+ZIP_DO_compressed_size);
 	}
-	if (info.uncompressed_size != le_uint32_read(buf+ZIP_DO_uncompressed_size)) {
+	if (info.uncompressed_size != le_uint32_read(buf+ZIP_DO_uncompressed_size)
+		// allow a 0 size, GNU unzip also allow it
+		&& 0 != le_uint32_read(buf+ZIP_DO_uncompressed_size)) {
 		throw error_invalid() << "Invalid uncompressed size in data descriptor " << info.uncompressed_size << "/" << le_uint32_read(buf+ZIP_DO_uncompressed_size);
 	}
 }
@@ -545,33 +554,6 @@ void zip_entry::unload()
 	data = 0;
 }
 
-/** Skip local file header and data. */
-void zip::skip_local(const unsigned char* buf, FILE* f)
-{
-	unsigned local_extra_field_length = le_uint16_read(buf+ZIP_LO_extra_field_length);
-	unsigned filename_length = le_uint16_read(buf+ZIP_LO_filename_length);
-	unsigned compressed_size = le_uint32_read(buf+ZIP_LO_compressed_size);
-	
-	// skip filename and extra field
-	if (fseek(f, filename_length + local_extra_field_length, SEEK_CUR) != 0) {
-		throw error_invalid() << "Failed seek";
-	}
-
-	// directory don't have data
-	if (compressed_size) {
-		if (fseek(f, compressed_size, SEEK_CUR) != 0) {
-			throw error_invalid() << "Failed seek";
-		}
-	}
-
-	// data descriptor
-	if ((le_uint16_read(buf+ZIP_LO_general_purpose_bit_flag) & ZIP_GEN_FLAGS_DEFLATE_ZERO) != 0) {
-		if (fseek(f, ZIP_DO_FIXED, SEEK_CUR) != 0) {
-			throw error_invalid() << "Failed seek";
-		}
-	}
-}
-
 /**
  * Load local file header.
  * \param buf Fixed size local header.
@@ -618,15 +600,24 @@ void zip_entry::load_local(const unsigned char* buf, FILE* f, unsigned size)
 	// load the data descriptor
 	if ((le_uint16_read(buf+ZIP_LO_general_purpose_bit_flag) & ZIP_GEN_FLAGS_DEFLATE_ZERO) != 0) {
 		unsigned char data_desc[ZIP_DO_FIXED];
+		unsigned offset;
 
-		if (size < ZIP_DO_FIXED) {
+		// handle the case of the ZIP_DO_header_signature missing
+		if (size == ZIP_DO_FIXED - 4) {
+			le_uint32_write(data_desc+ZIP_DO_header_signature, 0x08074b50);
+			offset = ZIP_DO_crc32;
+		} else {
+			offset = 0;
+		}
+
+		if (size < ZIP_DO_FIXED - offset) {
 			throw error_invalid() << "Overflow of data descriptor";
 		}
-		size -= ZIP_DO_FIXED;
 
-		if (fread(data_desc, ZIP_DO_FIXED, 1, f) != 1) {
+		if (fread(data_desc + offset, ZIP_DO_FIXED - offset, 1, f) != 1) {
 			throw error() << "Failed read";
 		}
+		size -= ZIP_DO_FIXED - offset;
 
 		check_descriptor(data_desc);
 	}
@@ -690,11 +681,11 @@ void zip_entry::save_local(FILE* f)
  * \param buf Fixed size cent dir.
  * \param f File seeked after the fixed size cent dir.
  */
-void zip_entry::load_cent(const unsigned char* buf, unsigned& skip)
+void zip_entry::load_cent(const unsigned char* buf, unsigned buf_size, unsigned& skip)
 {
 	const unsigned char* o_buf = buf;
 
-	check_cent(buf);
+	check_cent(buf, buf_size);
 
 	// read header
 	info.version_made_by = le_uint8_read(buf+ZIP_CO_version_made_by);
@@ -716,6 +707,14 @@ void zip_entry::load_cent(const unsigned char* buf, unsigned& skip)
 	info.relative_offset_of_local_header = le_uint32_read(buf+ZIP_CO_relative_offset_of_local_header);
 	buf += ZIP_CO_FIXED;
 
+	if (buf_size < info.filename_length
+		|| buf_size < info.central_extra_field_length
+		|| buf_size < info.file_comment_length
+		|| buf_size < ZIP_CO_FIXED + info.filename_length + info.central_extra_field_length + info.file_comment_length
+	) {
+		throw error_invalid() << "Invalid central directory data";
+	}
+
 	// read filename
 	data_free(file_name);
 	file_name = data_alloc(info.filename_length);
@@ -864,7 +863,7 @@ void zip::open()
 
 			unsigned skip = 0;
 			try {
-				i->load_cent(data + data_pos, skip);
+				i->load_cent(data + data_pos, data_size - data_pos, skip);
 			} catch (...) {
 				map.erase(i);
 				throw;
